import unittest
from typing import Any
from typing_extensions import TypedDict

from pydantic import ValidationError
from pydantic_yaml import parse_yaml_raw_as

from expedantic import ConfigBase


class PythonTypedDict(TypedDict):
    name: str
    age: int


class Config(ConfigBase):
    optional: int | None = 1
    dictionary: dict[str, Any] = {}
    typed_dictionary: dict[int, str] = {}
    untyped_list: list = []
    typed_list: list[int] = []
    python_typed_dict: PythonTypedDict = {"name": "Maria", "age": 30}


class TestCompoundTypes(unittest.TestCase):
    def test_optional(self):
        config = parse_yaml_raw_as(Config, "optional: null")
        self.assertEqual(config.optional, None)

    def test_dictionary(self):
        config = parse_yaml_raw_as(
            Config, "dictionary: {name: Maria, job: Developer, age: 30}"
        )
        self.assertEqual(config.dictionary["name"], "Maria")
        self.assertEqual(config.dictionary["job"], "Developer")
        self.assertEqual(config.dictionary["age"], 30)

    def test_typed_dictionary(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "typed_dictionary: {name: Maria}")

        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "typed_dictionary: {1: 20}")

        config = parse_yaml_raw_as(Config, "typed_dictionary: {1: Maria}")
        self.assertEqual(config.typed_dictionary[1], "Maria")

    def test_untyped_list(self):
        config = parse_yaml_raw_as(Config, "untyped_list: [1, 'Maria', 3.5, [1, 2]]")
        self.assertEqual(config.untyped_list, [1, "Maria", 3.5, [1, 2]])

    def test_typed_list(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "typed_list: [3, 4, 7.2]")

        config = parse_yaml_raw_as(Config, "typed_list: [0, 1, 2, 3]")
        self.assertEqual(config.typed_list, [0, 1, 2, 3])

    def test_python_typed_dict(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(
                Config,
                "python_typed_dict: {'name': Maria, 'age': 30, 'job': Developer}",
            )

        config = parse_yaml_raw_as(
            Config, "python_typed_dict: {'name': Tom, 'age': 17}"
        )
        self.assertEqual(config.python_typed_dict, {'name': 'Tom', 'age': 17})


if __name__ == "__main__":
    unittest.main()



import inspect
from typing import Any, Type

from tap.utils import type_to_str, get_literals


EMPTY = inspect.Parameter.empty
POS_ONLY = inspect.Parameter.POSITIONAL_ONLY


def get_kwargs(cls: Type):
    signature = inspect.signature(cls)
    kw_args = {
        k: (
            v.annotation if v.annotation != EMPTY else Any,
            v.default if v.default != EMPTY else ...,
        )
        for k, v in signature.parameters.items()
        if v.kind != POS_ONLY
    }
    return kw_args


def flatten_dict(d: dict[str, Any], parent_key="", sep=".") -> dict[str, Any]:
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



import unittest
from typing import Any, Literal

from expedantic import ConfigBase, Field


class TestInnerConfig(ConfigBase):
    a: int = -1
    b: float = -2.0


class TestConfig(ConfigBase):
    a: int = 1
    b: float = 2.0
    c: str = "3.0"
    d: list[float] = Field(default_factory=lambda: [1.0, 2.0, 3.0])
    e: TestInnerConfig = TestInnerConfig()
    f: Any = "asdf"
    g: Literal["A", "B", "C"]
    h: list[int]
    file: str


class TestBase(ConfigBase):
    a: int = 1
    b: float = 2.0


class TestChild(TestBase):
    c: str = "c"


# class TestSaveLoad(unittest.TestCase):
# def test_save(self):
# TestConfig.model_dump()


if __name__ == "__main__":
    # t = TestConfig()
    # tapified = TestConfig.tapify()
    c = TestConfig.parse_args()
    # TestConfig.generate_schema("test_schema.json")

    print(c)



import argparse
import inspect
import json

from abc import ABC
from collections.abc import Mapping
from io import IOBase
from pathlib import Path
from typing import Any, Type, Literal, get_origin, get_args

import pydantic
import pydantic_yaml

# from ruamel.yaml import YAML
from ccorp.ruamel.yaml.include import YAML

pydantic_yaml.loader.YAML = YAML

from . import utils


class NOT_PROVIDED_CLASS:
    def __repr__(self) -> str:
        return "NOT_PROVIDED"


_NOT_PROVIDED = NOT_PROVIDED_CLASS()


class ConfigBase(pydantic.BaseModel, Mapping, ABC):
    model_config = pydantic.ConfigDict(
        extra="forbid", protected_namespaces=("model_", "config_base_")
    )

    def flatten(self):
        return utils.flatten_dict(self.model_dump())

    def compatible_args(self, cls: Type, *exclusive_keys: str):
        fields = self.model_dump()
        args = utils.get_kwargs(cls)
        arg_keys = set(args.keys())
        arg_keys -= set(exclusive_keys)
        return {k: v for k, v in fields.items() if k in arg_keys}

    def save_as_yaml(
        self,
        file: Path | str | IOBase,
        default_flow_style: bool | None = False,
        indent: int | None = None,
        map_indent: int | None = None,
        sequence_indent: int | None = None,
        sequence_dash_offset: int | None = None,
        custom_yaml_writer: YAML | None = None,
        **json_kwargs,
    ):
        """Write a YAML file representation of the model.

        Parameters
        ----------
        file : Path or str or IOBase
            The file path or stream to write to.
        model : BaseModel
            The model to write.
        default_flow_style : bool
            Whether to use "flow style" (more human-readable).
            https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
        indent : None or int
            General indent value. Leave as None for the default.
        map_indent, sequence_indent, sequence_dash_offset : None or int
            More specific indent values.
        custom_yaml_writer : None or YAML
            An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
            The above options will be set on it, if given.
        json_kwargs : Any
            Keyword arguments to pass `model.json()`.

        Notes
        -----
        This currently uses JSON dumping as an intermediary.
        This means that you can use `json_encoders` in your model.
        """
        pydantic_yaml.to_yaml_file(
            file,
            self,
            default_flow_style=default_flow_style,
            indent=indent,
            map_indent=map_indent,
            sequence_indent=sequence_indent,
            sequence_dash_offset=sequence_dash_offset,
            custom_yaml_writer=custom_yaml_writer,
            **json_kwargs,
        )

    @classmethod
    def generate_schema(cls, path: str | Path):
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        json.dump(cls.model_json_schema(), path.open("w"), indent=2)
        print(f"JSON Schema for {cls.__name__} is generated at {path}")

    @classmethod
    def load_from_yaml(cls, path: Path | str):
        path = Path(path)
        return pydantic_yaml.parse_yaml_file_as(cls, path)

    @classmethod
    def parse_args(cls, require_default_file: bool = False, args=None, sep="."):
        def _parse_params(
            parser: argparse.ArgumentParser,
            cls: Type[ConfigBase],
            parent_key: str = "",
            sep=sep,
        ):
            for name, field_info in cls.model_fields.items():
                if parent_key:
                    name = f"{parent_key}{sep}{name}"

                tp = field_info.annotation
                assert tp is not None
                origin = get_origin(tp)

                if not origin and inspect.isclass(tp) and issubclass(tp, ConfigBase):
                    _parse_params(parser, tp, name)
                    continue

                names = [f"--{name}"]
                kwargs = {}
                kwargs["type"] = tp

                annot_repr = str(tp) if origin else tp.__name__

                if tp is Any:
                    kwargs["type"] = str
                elif origin is Literal:
                    var_type, literals = utils.get_literals(tp, name)
                    kwargs["type"] = var_type
                    kwargs["choices"] = literals
                    annot_repr = annot_repr.replace("typing.", "")
                elif origin in {list, set, tuple}:
                    kwargs["nargs"] = "*"
                    args = get_args(tp)
                    if len(args) == 0:
                        kwargs["type"] = str
                    else:
                        kwargs["type"] = args[0]

                if field_info.is_required():
                    req_repr = "required"
                    if not require_default_file:
                        kwargs["default"] = _NOT_PROVIDED
                else:
                    default_value = field_info.get_default(call_default_factory=True)
                    req_repr = f"default={default_value}"
                    if not require_default_file:
                        kwargs["default"] = default_value

                kwargs["help"] = f"({annot_repr}, {req_repr})"

                if require_default_file:
                    kwargs["default"] = _NOT_PROVIDED

                parser.add_argument(*names, **kwargs)

        parser = argparse.ArgumentParser()
        if require_default_file:
            parser.add_argument("_config_file_path", type=Path, help="")
        _parse_params(parser, cls)
        args = parser.parse_args(args=args)

        if require_default_file:
            with open(args._config_file_path, "r") as f:
                file_dict = YAML().load(f)
            args.__dict__.pop("_config_file_path")
        else:
            file_dict = {}

        nested_args_dict = {}
        for k, v in vars(args).items():
            if v is _NOT_PROVIDED:
                continue
            keys = k.split(sep)
            current_level = nested_args_dict
            for key in keys[:-1]:
                current_level = current_level.setdefault(key, {})
            current_level[keys[-1]] = v

        def update_recursively(d: dict, other: dict):
            for k, v in other.items():
                if isinstance(v, dict):
                    d[k] = update_recursively(d.get(k, {}), v)
                else:
                    d[k] = v
            return d

        file_dict = update_recursively(file_dict, nested_args_dict)
        instance = cls.model_validate(file_dict)

        return instance

    def __getitem__(self, key: str):
        if key not in self.model_fields:
            raise KeyError(f"Key '{key}' not found.")
        return self.__getattribute__(key)

    def __len__(self):
        return len(self.model_fields)

    def keys(self):
        return self.model_fields.keys()



from setuptools import setup, find_packages


setup(
    name="expedantic",
    version="0.1.3",
    packages=["expedantic", "ccorp.ruamel.yaml.include"],
    package_dir={
        # "expedantic": "src",
        "expedantic": "src/expedantic",
        "ccorp.ruamel.yaml.include": "./submodules/ccorp_yaml_include/ccorp/ruamel/yaml/include",
    },
    python_requires=">=3.10",
    install_requires=[
        "pydantic >= 2.6.3",
        "pydantic_yaml >= 1.2.1",
        "typed-argument-parser",
    ],
    author="rnilva",
    author_email="rnilva849@gmail.com",
    description="",
    license="MIT",
    keywords="",
    url="https://github.com/rnilva/expedantic",
)



# import ruamel.yaml

# yaml = ruamel.yaml.YAML()

# yaml.plug_ins

# from ccorp.ruamel.yaml.include import YAML

# yaml = YAML(typ="unsafe", pure=True)

# from expedantic.

# from ccorp.ruamel.yaml.include import
import expedantic

import pydantic

from ccorp.ruamel.yaml.include import YAML

# yaml = YAML()
